﻿using System;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using TSLab.Script.Helpers;

namespace TSLab.Script.Handlers
{
    public abstract class BasePeriodIndicatorHandler
    {
        private int m_period = 1;

        [HandlerParameter(true, "20", Min = "10", Max = "100", Step = "5", EditorMin = "1")]
        public int Period
        {
            get { return m_period; }
            set { m_period = Math.Max(value, 1); }
        }
    }

    public abstract class BaseRSI<THandler> : DoubleStreamAndValuesHandlerWithPeriod
        where THandler : DoubleStreamAndValuesHandlerWithPeriod, new()
    {
        private double m_lastSource;
        private THandler m_uHandler;
        private THandler m_dHandler;

        public override bool IsGapTolerant
        {
            get { return false; }
        }

        protected override void InitExecuteContext()
        {
            m_lastSource = 0;
            m_uHandler = new THandler { Context = Context, Period = Period };
            m_dHandler = new THandler { Context = Context, Period = Period };
        }

        protected override void ClearExecuteContext()
        {
            m_lastSource = 0;
            m_uHandler = null;
            m_dHandler = null;
        }

        protected override void InitForGap()
        {
            var firstIndex = Math.Max(m_executeContext.LastIndex + 1, m_executeContext.Index - GapTolerancePeriodMultiplier * Period);
            for (var i = firstIndex; i < m_executeContext.Index; i++)
            {
                var source = m_executeContext.GetSourceForGap(i);
                double uEma, dEma;
                Precalc(source, i, out uEma, out dEma);
            }
        }

        protected override double Execute()
        {
            double uEma, dEma;
            Precalc(m_executeContext.Source, m_executeContext.Index, out uEma, out dEma);
            double result;

            if (dEma == 0.0)
                result = 100;
            else if (uEma / dEma == 1.0)
                result = 0;
            else
                result = 100 - 100 / (1 + uEma / dEma);

            return result;
        }

        private void Precalc(double source, int index, out double uEma, out double dEma)
        {
            double u = 0, d = 0;
            if (index > 0)
            {
                if (m_lastSource < source)
                    u = source - m_lastSource;
                else if (m_lastSource > source)
                    d = m_lastSource - source;
            }
            m_lastSource = source;
            uEma = m_uHandler.Execute(u, index);
            dEma = m_dHandler.Execute(d, index);
        }
    }

    [HandlerCategory(HandlerCategories.Indicators)]
    public sealed class RSI : BaseRSI<EMA>
    {
        public override IList<double> Execute(IList<double> source)
        {
            var result = Series.RSI(source, Period);
            return result;
        }
    }

    //[HandlerName("Cutler's RSI")]
    [HandlerCategory(HandlerCategories.Indicators)]
    [SuppressMessage("StyleCopPlus.StyleCopPlusRules", "SP0100:AdvancedNamingRules", Justification = "Reviewed. Suppression is OK here.")]
    public sealed class CuttlerRSI : BaseRSI<SMA>
    {
        public override IList<double> Execute(IList<double> source)
        {
            var result = Series.CuttlerRSI(source, Period);
            return result;
        }
    }

    public abstract class SumBase : DoubleStreamAndValuesHandlerWithPeriod
    {
        private readonly bool m_isSma;
        private Queue<double> m_queue;
        private double m_sum;

        protected SumBase(bool isSma)
        {
            m_isSma = isSma;
        }

        public override bool IsGapTolerant
        {
            get { return true; }
        }

        protected override void InitExecuteContext()
        {
            if (IsSimple)
                return;

            m_queue = new Queue<double>(Period);
            m_sum = 0;
        }

        protected override void ClearExecuteContext()
        {
            m_queue = null;
            m_sum = 0;
        }

        protected override void InitForGap()
        {
            if (IsSimple)
                return;

            var firstIndex = Math.Max(m_executeContext.LastIndex + 1, m_executeContext.Index - Period + 1);
            for (var i = firstIndex; i < m_executeContext.Index; i++)
            {
                var source = m_executeContext.GetSourceForGap(i);
                var result = Calc(source);
            }
        }

        protected override double Execute()
        {
            if (IsSimple)
                return m_executeContext.Source;

            var result = Calc(m_executeContext.Source);
            return result;
        }

        private double Calc(double source)
        {
            if (m_queue.Count == Period)
                m_sum -= m_queue.Dequeue();

            m_queue.Enqueue(source);
            m_sum += source;
            return m_isSma ? m_sum / m_queue.Count : m_sum;
        }
    }

    [HandlerCategory(HandlerCategories.Indicators)]
    public sealed class SummFor : SumBase
    {
        public SummFor()
            : base(false)
        {
        }

        public override IList<double> Execute(IList<double> source)
        {
            var result = Series.SummFor(source, Period);
            return result;
        }
    }

    [HandlerCategory(HandlerCategories.Indicators)]
    public sealed class SMA : SumBase
    {
        public SMA()
            : base(true)
        {
        }

        public override IList<double> Execute(IList<double> source)
        {
            var result = Series.SMA(source, Period);
            return result;
        }
    }

    public abstract class StDevVolatility : DoubleStreamAndValuesHandlerWithPeriod
    {
        protected ShrinkedList<double> m_source;
        protected ShrinkedList<double> m_smaValues;
        private SMA m_smaHandler;

        public override bool IsGapTolerant
        {
            get { return true; }
        }

        protected override void InitExecuteContext()
        {
            if (IsSimple)
                return;

            m_source = new ShrinkedList<double>(Period);
            m_smaValues = new ShrinkedList<double>(Period);
            m_smaHandler = new SMA { Context = Context, Period = Period };
        }

        protected override void ClearExecuteContext()
        {
            m_source = null;
            m_smaValues = null;
            m_smaHandler = null;
        }

        protected override void InitForGap()
        {
            if (IsSimple)
                return;

            var firstIndex = Math.Max(m_executeContext.LastIndex + 1, m_executeContext.Index - Period + 1);
            for (var i = firstIndex; i < m_executeContext.Index; i++)
            {
                var source = m_executeContext.GetSourceForGap(i);
                m_source.Add(source);
                m_smaValues.Add(m_smaHandler.Execute(source, i));
            }
        }

        protected override double Execute()
        {
            if (IsSimple)
                return 0;

            m_source.Add(m_executeContext.Source);
            m_smaValues.Add(m_smaHandler.Execute(m_executeContext.Source, m_executeContext.Index));
            var result = Calc();
            return result;
        }

        protected abstract double Calc();
    }

    [HandlerCategory(HandlerCategories.Indicators)]
    public sealed class StDev : StDevVolatility
    {
        public override IList<double> Execute(IList<double> source)
        {
            var result = Series.StDev(source, Period);
            return result;
        }

        protected override double Calc()
        {
            var result = Indicators.StDev(m_source, m_smaValues, m_source.Count - 1, Period);
            return result;
        }
    }

    [HandlerCategory(HandlerCategories.Indicators)]
    public sealed class Volatility : StDevVolatility
    {
        public override IList<double> Execute(IList<double> source)
        {
            var result = Series.Volatility(source, Period);
            return result;
        }

        protected override double Calc()
        {
            var result = Indicators.Volatility(m_source, m_smaValues, m_source.Count - 1, Period);
            return result;
        }
    }

    [HandlerCategory(HandlerCategories.Indicators)]
    public sealed class EMA : DoubleStreamAndValuesHandlerWithPeriod
    {
        private SMA m_sma;
        private double m_k;
        private double m_lastResult;

        public override bool IsGapTolerant
        {
            get { return IsSimple; }
        }

        public override IList<double> Execute(IList<double> source)
        {
            var result = Series.EMA(source, Period);
            return result;
        }

        protected override void InitExecuteContext()
        {
            if (IsSimple)
                return;

            m_sma = new SMA { Context = Context, Period = Period };
            m_k = 2D / (1D + Period);
            m_lastResult = 0;
        }

        protected override void ClearExecuteContext()
        {
            m_sma = null;
            m_lastResult = 0;
        }

        protected override void InitForGap()
        {
            if (IsSimple)
                return;

            var firstIndex = Math.Max(m_executeContext.LastIndex + 1, m_executeContext.Index - GapTolerancePeriodMultiplier * Period);
            for (var i = firstIndex; i < m_executeContext.Index; i++)
            {
                var source = m_executeContext.GetSourceForGap(i);
                var result = Calc(source, i);
            }
        }

        protected override double Execute()
        {
            if (IsSimple)
                return m_executeContext.Source;

            var result = Calc(m_executeContext.Source, m_executeContext.Index);
            return result;
        }

        private double Calc(double source, int index)
        {
            if (index < Period)
                m_lastResult = m_sma.Execute(source, index);
            else
                m_lastResult = m_k * (source - m_lastResult) + m_lastResult;

            return m_lastResult;
        }
    }

    //[HandlerName("TR")]
    [HandlerCategory(HandlerCategories.Indicators)]
    public sealed class TrueRange : IBar2DoubleHandler
    {
        public IList<double> Execute(ISecurity source)
        {
            return Series.TrueRange(source.Bars);
        }
    }

    //[HandlerName("ATR")]
    [HandlerCategory(HandlerCategories.Indicators)]
    public sealed class AverageTrueRange : BasePeriodIndicatorHandler, IBar2DoubleHandler
    {
        public IList<double> Execute(ISecurity source)
        {
            return Series.AverageTrueRange(source.Bars, Period);
        }
    }

    [HandlerCategory(HandlerCategories.Indicators)]
    public sealed class CCI : BasePeriodIndicatorHandler, IBar2DoubleHandler
    {
        public IList<double> Execute(ISecurity source)
        {
            return Series.CCI(source.Bars, Period);
        }
    }

    //[HandlerName("Typical Price")]
    [HandlerCategory(HandlerCategories.Indicators)]
    public sealed class TypicalPrice : IBar2DoubleHandler
    {
        public IList<double> Execute(ISecurity source)
        {
            return Series.TypicalPrice(source.Bars);
        }
    }

    public abstract class BollingerBandsBase : DoubleStreamAndValuesHandlerWithPeriod
    {
        private readonly bool m_isTopLine;
        private SMA m_sma;
        private StDev m_stDev;

        protected BollingerBandsBase(bool isTopLine)
        {
            m_isTopLine = isTopLine;
        }

        [HandlerParameter(true, "2", Min = "0.5", Max = "3", Step = "0.5")]
        public double Coef { get; set; }

        public override bool IsGapTolerant
        {
            get { return true; }
        }

        public override IList<double> Execute(IList<double> source)
        {
            var result = Series.BollingerBands(source, Period, Coef, m_isTopLine);
            return result;
        }

        protected override void InitExecuteContext()
        {
            if (IsSimple)
                return;

            m_sma = new SMA { Context = Context, Period = Period };
            m_stDev = new StDev { Context = Context, Period = Period };
        }

        protected override void ClearExecuteContext()
        {
            m_sma = null;
            m_stDev = null;
        }

        protected override void InitForGap()
        {
            if (IsSimple)
                return;

            var firstIndex = Math.Max(m_executeContext.LastIndex + 1, m_executeContext.Index - Period + 1);
            for (var i = firstIndex; i < m_executeContext.Index; i++)
            {
                var source = m_executeContext.GetSourceForGap(i);
                var sma = m_sma.Execute(source, i);
                var stDev = m_stDev.Execute(source, i);
            }
        }

        protected override double Execute()
        {
            if (IsSimple)
                return m_executeContext.Source;

            var sma = m_sma.Execute(m_executeContext.Source, m_executeContext.Index);
            var stDev = m_stDev.Execute(m_executeContext.Source, m_executeContext.Index);
            var sign = m_isTopLine ? 1 : -1;
            var result = sma + sign * Coef * stDev;
            return result;
        }
    }

    //[HandlerName("Bollinger Bands (+)")]
    [HandlerCategory(HandlerCategories.Indicators)]
    public sealed class BollingerBands1 : BollingerBandsBase
    {
        public BollingerBands1()
            : base(true)
        {
        }
    }

    //[HandlerName("Bollinger Bands (-)")]
    [HandlerCategory(HandlerCategories.Indicators)]
    public sealed class BollingerBands2 : BollingerBandsBase
    {
        public BollingerBands2()
            : base(false)
        {
        }
    }

    [HandlerDecimals(2)]
    [HandlerCategory(HandlerCategories.Indicators)]
    public class Relative : IDouble2DoubleHandler
    {
        public IList<double> Execute(IList<double> source)
        {
            double v = source.Count == 0 ? 0 : source[0];
// ReSharper disable CompareOfFloatsByEqualityOperator
            if (v == 0)
                v = double.MaxValue;
            // ReSharper restore CompareOfFloatsByEqualityOperator
            IList<double> list = new List<double>(source.Count);
            foreach (double t in source)
                list.Add((t - v) / v * 100);
            return list;
        }
    }

    public abstract class HighestLowest : DoubleStreamAndValuesHandlerWithPeriod
    {
        protected ShrinkedList<double> m_source;

        public override bool IsGapTolerant
        {
            get { return true; }
        }

        protected override void InitExecuteContext()
        {
            if (IsSimple)
                return;

            m_source = new ShrinkedList<double>(Period);
        }

        protected override void ClearExecuteContext()
        {
            m_source = null;
        }

        protected override void InitForGap()
        {
            if (IsSimple)
                return;

            var firstIndex = Math.Max(m_executeContext.LastIndex + 1, m_executeContext.Index - Period + 1);
            for (var i = firstIndex; i < m_executeContext.Index; i++)
                m_source.Add(m_executeContext.GetSourceForGap(i));
        }

        protected override double Execute()
        {
            if (IsSimple)
                return m_executeContext.Source;

            m_source.Add(m_executeContext.Source);
            var result = Calc();
            return result;
        }

        protected abstract double Calc();
    }

    [HandlerCategory(HandlerCategories.Indicators)]
    public sealed class Highest : HighestLowest
    {
        public override IList<double> Execute(IList<double> source)
        {
            var result = Series.Highest(source, Period);
            return result;
        }

        protected override double Calc()
        {
            var result = Indicators.Highest(m_source, m_source.Count - 1, Period);
            return result;
        }
    }

    [HandlerCategory(HandlerCategories.Indicators)]
    public sealed class Lowest : HighestLowest
    {
        public override IList<double> Execute(IList<double> source)
        {
            var result = Series.Lowest(source, Period);
            return result;
        }

        protected override double Calc()
        {
            var result = Indicators.Lowest(m_source, m_source.Count - 1, Period);
            return result;
        }
    }

    [HandlerCategory(HandlerCategories.TradeMath)]
    //[HandlerName("Multiply by Coef", Language = "en-US")]
    //[HandlerName("Умножить на число", Language = "ru-RU")]
    public class Multiply : IDouble2DoubleHandler
    {
        [HandlerParameter(true, "2", Min = "0.5", Max = "5", Step = "0.5")]
        public double Coef
        {
            get;
            set;
        }

        public IList<double> Execute(IList<double> bars)
        {
            IList<double> list = new List<double>(bars.Count);
            foreach (double t in bars)
                list.Add(t * Coef);
            return list;
        }
    }

    [HandlerCategory(HandlerCategories.TradeMath)]
    public class Ln : IDouble2DoubleHandler
    {
        protected double m_mult = 1, m_add = 0;

        [HandlerParameter(true, Default = "1")]
        public double Mult
        {
            get { return m_mult; }
            set { m_mult = value; }
        }

        [HandlerParameter(true, Default = "0")]
        public double Add
        {
            get { return m_add; }
            set { m_add = value; }
        }

        public IList<double> Execute(IList<double> bars)
        {
            IList<double> list = new List<double>(bars.Count);
            for (int j = 0; j < bars.Count; j++)
            {
                double res = m_mult * Math.Log(bars[j]) + m_add;
                list.Add(res);
            }
            return list;
        }
    }
}
