﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Windows.Media;
using TSLab.DataSource;
using TSLab.Script.GraphPane;
using TSLab.Script.Optimization;

namespace TSLab.Script.Handlers
{
    [HandlerCategory(BlockCategories.GraphPaneHandler)]
    [InputsCount(2)]
    [Input(0, TemplateTypes.GRAPHPANE, Name = "TemplateLibrary.Pane")]
    [Input(1, TemplateTypes.SECURITY, Name = "Security")]
    public sealed partial class InteractiveLineGen : IInteractiveLineGen
    {
        private static readonly IList<double> s_defaultResult0 = new double[0];
        private static readonly IList<double> s_defaultResult1 = new[] { double.NaN };
        private IInteractiveLine m_interactiveLine;
        private IReadOnlyList<IDataBar> m_bars;

        public IContext Context { get; set; }

        public string VariableId { get; set; }

        [HandlerParameter(Default = "RIGHT", IsShown = false, NotOptimized = true)]
        public PaneSides PaneSide { get; set; }

        [HandlerParameter(Default = "#ff0000", IsShown = false, NotOptimized = true, Editor = "StringToColorTemplate")]
        public string Color { get; set; }

        [HandlerParameter(Default = "Infinite", IsShown = false, NotOptimized = true)]
        public InteractiveLineMode Mode { get; set; }

        [HandlerParameter(Name = "FirstDateTime", Default = "0001-01-01T00:00:00.0000000", IsShown = false)]
        public DateTimeOptimProperty FirstDateTime { get; set; }

        [HandlerParameter(Name = "FirstValue", Default = "0", IsShown = false)]
        public OptimProperty FirstValue { get; set; }

        [HandlerParameter(Name = "SecondDateTime", Default = "9999-12-31T23:59:59.9999999", IsShown = false)]
        public DateTimeOptimProperty SecondDateTime { get; set; }

        [HandlerParameter(Name = "SecondValue", Default = "0", IsShown = false)]
        public OptimProperty SecondValue { get; set; }

        [HandlerParameter(Default = "true", IsShown = false, NotOptimized = true)]
        public bool IsNeedRecalculate { get; set; }

        public IList<double> Execute(IGraphPane pane, ISecurity security)
        {
            if (pane == null)
                throw new ArgumentNullException(nameof(pane));

            if (security == null)
                throw new ArgumentNullException(nameof(security));

            var id = typeof(IInteractiveLineGen).Name + "." + VariableId;
            var container = (NotClearableContainer<InteractiveLineGen>)Context.LoadObject(id);
            container?.Content.Unsubscribe();
            var result = GetLine(pane, m_bars = security.Bars);
            Subscribe();
            Context.StoreObject(id, new NotClearableContainer<InteractiveLineGen>(this));
            return result;
        }

        private IList<double> GetLine(IGraphPane pane, IReadOnlyList<IDataBar> bars)
        {
            m_interactiveLine = null;
            if (bars == null || bars.Count == 0)
                return s_defaultResult0;

            if (bars.Count == 1)
                return s_defaultResult1;

            var interactiveLine = GetInteractiveLine(pane, bars);
            var firstPosition = interactiveLine.FirstPoint.MarketPosition;
            var secondPosition = interactiveLine.SecondPoint.MarketPosition;

            if (firstPosition.X == secondPosition.X)
                return new ConstList(double.NaN, bars.Count);

            var firstBarIndex = GetBarIndex(bars, firstPosition.X);
            var secondBarIndex = GetBarIndex(bars, secondPosition.X);

            if (firstBarIndex == null || secondBarIndex == null || firstBarIndex == secondBarIndex)
                return new ConstList(double.NaN, bars.Count);

            var a = (secondPosition.Y - firstPosition.Y) / (secondBarIndex.Value - firstBarIndex.Value);
            var b = firstPosition.Y - a * firstBarIndex.Value;

            int minIndex;
            int maxIndex;

            switch (Mode)
            {
                case InteractiveLineMode.Finite:
                    if (firstBarIndex < secondBarIndex)
                    {
                        minIndex = firstBarIndex.Value;
                        maxIndex = secondBarIndex.Value;
                    }
                    else
                    {
                        minIndex = secondBarIndex.Value;
                        maxIndex = firstBarIndex.Value;
                    }
                    break;

                case InteractiveLineMode.Infinite:
                    minIndex = 0;
                    maxIndex = bars.Count - 1;
                    break;

                case InteractiveLineMode.Ray:
                    if (firstBarIndex < secondBarIndex)
                    {
                        minIndex = firstBarIndex.Value;
                        maxIndex = bars.Count - 1;
                    }
                    else
                    {
                        minIndex = 0;
                        maxIndex = firstBarIndex.Value;
                    }
                    break;

                default:
                    throw new InvalidOperationException();
            }
            return a == 0 ? (IList<double>)new ConstList(b, bars.Count, minIndex, maxIndex) : new LineList(a, b, bars.Count, minIndex, maxIndex);
        }

        private IInteractiveLine GetInteractiveLine(IGraphPane pane, IReadOnlyList<IDataBar> bars)
        {
            var minDate = bars.First().Date;
            var maxDate = bars.Last().Date;
            var interactiveLine = (IInteractiveLine)pane.GetInteractiveObject(VariableId);
            var mediaColor = (System.Windows.Media.Color)ColorConverter.ConvertFromString(Color);
            var intColor = mediaColor.ToInt();
            MarketPoint firstMarketPosition;
            MarketPoint secondMarketPosition;

            if (interactiveLine != null)
            {
                interactiveLine.ExecutionDataBars = bars;
                firstMarketPosition = interactiveLine.FirstPoint.MarketPosition;
                secondMarketPosition = interactiveLine.SecondPoint.MarketPosition;
                CorrectMarketPointsEx(ref firstMarketPosition, ref secondMarketPosition, minDate, maxDate);

                if (interactiveLine.PaneSides == PaneSide && interactiveLine.Color == intColor && interactiveLine.Mode == Mode)
                {
                    pane.AddUnremovableInteractiveObjectId(VariableId);
                    interactiveLine.FirstPoint.MarketPosition = firstMarketPosition;
                    interactiveLine.SecondPoint.MarketPosition = secondMarketPosition;
                    return m_interactiveLine = interactiveLine;
                }
                pane.RemoveInteractiveObject(VariableId);
            }
            else
            {
                firstMarketPosition = new MarketPoint(FirstDateTime.Value, FirstValue.Value);
                secondMarketPosition = new MarketPoint(SecondDateTime.Value, SecondValue.Value);
                CorrectMarketPointsEx(ref firstMarketPosition, ref secondMarketPosition, minDate, maxDate);
            }
            m_interactiveLine = pane.AddInteractiveLine(VariableId, PaneSide, false, intColor, Mode, firstMarketPosition, secondMarketPosition);
            m_interactiveLine.ExecutionDataBars = bars;
            return m_interactiveLine;
        }

        private void CorrectMarketPointsEx(ref MarketPoint firstMarketPosition, ref MarketPoint secondMarketPosition, DateTime minDate, DateTime maxDate)
        {
            CorrectMarketPoints(ref firstMarketPosition, ref secondMarketPosition, minDate, maxDate);
            SetValue(FirstDateTime, firstMarketPosition.X);
            SetValue(FirstValue, firstMarketPosition.Y);
            SetValue(SecondDateTime, secondMarketPosition.X);
            SetValue(SecondValue, secondMarketPosition.Y);
        }

        private static void CorrectMarketPoints(ref MarketPoint firstMarketPosition, ref MarketPoint secondMarketPosition, DateTime minDate, DateTime maxDate)
        {
            var firstDate = firstMarketPosition.X;
            var secondDate = secondMarketPosition.X;

            if (firstDate < secondDate)
            {
                if (firstDate < minDate || firstDate >= maxDate)
                    firstDate = minDate;

                if (secondDate <= minDate || secondDate > maxDate)
                    secondDate = maxDate;
            }
            else if (firstDate > secondDate)
            {
                if (secondDate < minDate || secondDate >= maxDate)
                    secondDate = minDate;

                if (firstDate <= minDate || firstDate > maxDate)
                    firstDate = maxDate;
            }
            else
            {
                if (firstDate < minDate)
                    firstDate = minDate;
                else if (firstDate > maxDate)
                    firstDate = maxDate;

                if (firstMarketPosition.X != firstDate)
                {
                    firstMarketPosition = new MarketPoint(firstDate, firstMarketPosition.Y);
                    secondMarketPosition = new MarketPoint(firstDate, secondMarketPosition.Y);
                }
                return;
            }
            if (firstMarketPosition.X != firstDate || secondMarketPosition.X != secondDate)
            {
                var a = (secondMarketPosition.Y - firstMarketPosition.Y) / (secondMarketPosition.X.Ticks - firstMarketPosition.X.Ticks);
                var b = firstMarketPosition.Y - a * firstMarketPosition.X.Ticks;

                if (firstMarketPosition.X != firstDate)
                    firstMarketPosition = new MarketPoint(firstDate, a * firstDate.Ticks + b);

                if (secondMarketPosition.X != secondDate)
                    secondMarketPosition = new MarketPoint(secondDate, a * secondDate.Ticks + b);
            }
        }

        private static int? GetBarIndex(IReadOnlyList<IDataBar> bars, DateTime date)
        {
            if (date < bars.First().Date || date > bars.Last().Date)
                return null;

            for (var i = 1; i < bars.Count; i++)
                if (date < bars[i].Date)
                    return i - 1;

            return bars.Count - 1;
        }

        private void Unsubscribe()
        {
            if (m_interactiveLine != null)
            {
                FirstUnsubscribe();
                SecondUnsubscribe();
            }
        }

        private void Subscribe()
        {
            if (m_interactiveLine != null)
            {
                FirstSubscribe();
                SecondSubscribe();
            }
        }

        private void FirstUnsubscribe()
        {
            FirstDateTime.PropertyChanged -= OnFirstDateTimePropertyChanged;
            FirstValue.PropertyChanged -= OnFirstValuePropertyChanged;
            m_interactiveLine.FirstPoint.PropertyChanged -= OnFirstPointPropertyChanged;
        }

        private void FirstSubscribe()
        {
            FirstDateTime.PropertyChanged += OnFirstDateTimePropertyChanged;
            FirstValue.PropertyChanged += OnFirstValuePropertyChanged;
            m_interactiveLine.FirstPoint.PropertyChanged += OnFirstPointPropertyChanged;
        }

        private void OnFirstDateTimePropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(IOptimPropertyBase.Value))
            {
                FirstUnsubscribe();
                var dateTimeOptimProperty = (DateTimeOptimProperty)sender;
                var firstMarketPosition = new MarketPoint(dateTimeOptimProperty.Value, m_interactiveLine.FirstPoint.MarketPosition.Y);
                var secondMarketPosition = m_interactiveLine.SecondPoint.MarketPosition;
                CorrectMarketPoints(ref firstMarketPosition, ref secondMarketPosition, m_bars.First().Date, m_bars.Last().Date);
                SetValue(dateTimeOptimProperty, firstMarketPosition.X);
                SetValue(FirstValue, firstMarketPosition.Y);
                m_interactiveLine.FirstPoint.MarketPosition = firstMarketPosition;
                FirstSubscribe();
            }
        }

        private void OnFirstValuePropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(IOptimPropertyBase.Value))
            {
                FirstUnsubscribe();
                var point = m_interactiveLine.FirstPoint;
                point.MarketPosition = new MarketPoint(point.MarketPosition.X, ((OptimProperty)sender).Value);
                FirstSubscribe();
            }
        }

        private void OnFirstPointPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            var interactivePoint = (IInteractivePoint)sender;
            if (e.PropertyName == nameof(IInteractivePoint.MarketPosition))
            {
                FirstUnsubscribe();
                var marketPosition = interactivePoint.MarketPosition;
                SetValue(FirstDateTime, marketPosition.X);
                SetValue(FirstValue, marketPosition.Y);
                FirstSubscribe();
            }
            if (IsNeedRecalculate && !interactivePoint.IsMoving && (e.PropertyName == nameof(IInteractivePoint.MarketPosition) || e.PropertyName == nameof(IInteractivePoint.IsMoving)))
                Context.Recalc();
        }

        private void SecondUnsubscribe()
        {
            SecondDateTime.PropertyChanged -= OnSecondDateTimePropertyChanged;
            SecondValue.PropertyChanged -= OnSecondValuePropertyChanged;
            m_interactiveLine.SecondPoint.PropertyChanged -= OnSecondPointPropertyChanged;
        }

        private void SecondSubscribe()
        {
            SecondDateTime.PropertyChanged += OnSecondDateTimePropertyChanged;
            SecondValue.PropertyChanged += OnSecondValuePropertyChanged;
            m_interactiveLine.SecondPoint.PropertyChanged += OnSecondPointPropertyChanged;
        }

        private void OnSecondDateTimePropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(IOptimPropertyBase.Value))
            {
                SecondUnsubscribe();
                var dateTimeOptimProperty = (DateTimeOptimProperty)sender;
                var firstMarketPosition = m_interactiveLine.FirstPoint.MarketPosition;
                var secondMarketPosition = new MarketPoint(dateTimeOptimProperty.Value, m_interactiveLine.SecondPoint.MarketPosition.Y);
                CorrectMarketPointsEx(ref firstMarketPosition, ref secondMarketPosition, m_bars.First().Date, m_bars.Last().Date);
                SetValue(dateTimeOptimProperty, secondMarketPosition.X);
                SetValue(SecondValue, secondMarketPosition.Y);
                m_interactiveLine.SecondPoint.MarketPosition = secondMarketPosition;
                SecondSubscribe();
            }
        }

        private void OnSecondValuePropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(IOptimPropertyBase.Value))
            {
                SecondUnsubscribe();
                var point = m_interactiveLine.SecondPoint;
                point.MarketPosition = new MarketPoint(point.MarketPosition.X, ((OptimProperty)sender).Value);
                SecondSubscribe();
            }
        }

        private void OnSecondPointPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            var interactivePoint = (IInteractivePoint)sender;
            if (e.PropertyName == nameof(IInteractivePoint.MarketPosition))
            {
                SecondUnsubscribe();
                var marketPosition = interactivePoint.MarketPosition;
                SetValue(SecondDateTime, marketPosition.X);
                SetValue(SecondValue, marketPosition.Y);
                SecondSubscribe();
            }
            if (IsNeedRecalculate && !interactivePoint.IsMoving && (e.PropertyName == nameof(IInteractivePoint.MarketPosition) || e.PropertyName == nameof(IInteractivePoint.IsMoving)))
                Context.Recalc();
        }

        private static void SetValue<TValue>(IOptimPropertyBase<TValue> optimPropertyBase, TValue value)
        {
            ((OptimDataBase)optimPropertyBase.Data).Value = optimPropertyBase.Value = value;
        }
    }
}
