﻿using System;
using System.Linq;
using TSLab.Script.Realtime;

// ReSharper disable UnusedMember.Global
// ReSharper disable MemberCanBePrivate.Global
// ReSharper disable UnusedAutoPropertyAccessor.Global
// ReSharper disable ClassNeverInstantiated.Global
namespace TSLab.Script.Handlers
{
    [HandlerCategory(HandlerCategories.Portfolio)]
    public class FreeMoney : IBar2ValueDoubleHandler
    {
        public double Execute(ISecurity source, int barNum)
        {
            var srt = source as ISecurityRt;
            if (srt != null)
            {
                return srt.CurrencyBalance;
            }
            double profit = source.InitDeposit;
            foreach (var pos in source.Positions.GetClosedOrActiveForBar(barNum))
            {
                if (pos.IsActiveForBar(barNum))
                {
                    profit -= pos.EntryPrice * pos.Shares * source.LotSize;
                }
                else
                {
                    profit += pos.Profit();
                }
            }
            return profit;
        }
    }

    [HandlerCategory(HandlerCategories.Portfolio)]
    public sealed class EstimatedMoney : IBar2ValueDoubleHandler
    {
        public double Execute(ISecurity source, int barNum)
        {
            var securityRt = source as ISecurityRt;
            return securityRt?.EstimatedBalance ?? source.InitDeposit + source.GetProfit(barNum);
        }
    }

    [HandlerCategory(HandlerCategories.Portfolio)]
    public class CurrentPosition : IBar2ValueDoubleHandler
    {
        public double Execute(ISecurity source, int barNum)
        {
            var srt = source as ISecurityRt;
            if (srt != null)
            {
                return srt.BalanceQuantity;
            }
            var activePos = source.Positions.GetActiveForBar(barNum);
            return activePos.Sum(pos => pos.Shares * (pos.IsLong ? 1 : -1));
        }
    }

    [HandlerCategory(HandlerCategories.Portfolio)]
    public sealed class WholeTimeProfit : IBar2ValueDoubleHandler
    {
        public double Execute(ISecurity source, int barNum)
        {
            return source.GetProfit(barNum);
        }
    }

    [HandlerCategory(HandlerCategories.Portfolio)]
    public class WholeDayProfit : IBar2ValueDoubleHandler
    {
        private int m_oldStartBarNum = -1;

        private DateTime m_oldStartDate;

        public double Execute(ISecurity source, int barNum)
        {
            return CalcProfit(source, barNum, 1, ref m_oldStartDate, ref m_oldStartBarNum);
        }

        public static double CalcProfit(ISecurity source, int barNum, int days, ref DateTime oldStartDate, ref int oldStartBarNum)
        {
            var dt = source.Bars[barNum].Date;
            var startDay = dt.Date;
            var endDay = startDay.AddDays(1);
            var startBarNum = barNum;
            if (days > 1)
            {
                startDay = startDay.AddDays(1 - days);
                if (startDay < source.Bars[0].Date)
                {
                    startDay = source.Bars[0].Date;
                    startBarNum = 0;
                }
            }
            if (startDay > oldStartDate)
            {
                oldStartDate = startDay;
                if (days > 1)
                {
                    for (int i = barNum; i >= 0; i--)
                    {
                        if (source.Bars[i].Date >= startDay) continue;
                        startBarNum = i + 1;
                        break;
                    }
                }
                oldStartBarNum = startBarNum;
            }
            double profit = 0;
            foreach (var pos in source.Positions)
            {
                if (pos.EntryBarNum <= barNum && (pos.IsActive || (pos.ExitBar.Date >= startDay && pos.EntryBar.Date < endDay)))
                {
                    profit += pos.GetProfit(barNum);
                    if (pos.EntryBarNum < oldStartBarNum)
                    {
                        profit -= pos.CurrentProfitByOpenPrice(oldStartBarNum);
                    }
                }
            }
            return profit;
        }
    }

    [HandlerCategory(HandlerCategories.Portfolio)]
    public class NumDaysProfit : BasePeriodIndicatorHandler, IBar2ValueDoubleHandler
    {
        private int m_oldDayStart = -1;

        private DateTime m_oldStartDay;

        public double Execute(ISecurity source, int barNum)
        {
            return WholeDayProfit.CalcProfit(source, barNum, Period, ref m_oldStartDay, ref m_oldDayStart);
        }
    }

    [HandlerCategory(HandlerCategories.Portfolio)]
    public class NumMinutesProfit : BasePeriodIndicatorHandler, IBar2ValueDoubleHandler
    {
        public double Execute(ISecurity source, int barNum)
        {
            var endDay = source.Bars[barNum].Date;
            var startDay = endDay.AddDays(-Period);
            return source.Positions.GetClosedOrActiveForBar(barNum)
                .Where(pos => pos.IsActive || (pos.ExitBar.Date >= startDay && pos.EntryBar.Date < endDay))
                .Sum(pos => pos.GetProfit(barNum));
        }
    }

    [HandlerCategory(HandlerCategories.Portfolio)]
    public class NumPositionsProfit : BasePeriodIndicatorHandler, IBar2ValueDoubleHandler
    {
        public double Execute(ISecurity source, int barNum)
        {
            double profit = 0;
            double trades = Period;
            foreach (var pos in source.Positions
                                    .Where(pos => !pos.IsActiveForBar(barNum))
                                    .OrderByDescending(pos => pos.ExitBarNum))
            {
                profit += pos.GetProfit(barNum);
                if (--trades <= 0)
                {
                    break;
                }
            }
            return profit;
        }
    }
}
