﻿using System;
using System.Linq;
using TSLab.DataSource;
using TSLab.Script.Realtime;

// ReSharper disable UnusedAutoPropertyAccessor.Global
// ReSharper disable MemberCanBePrivate.Global
// ReSharper disable UnusedMember.Global
// ReSharper disable ClassNeverInstantiated.Global
namespace TSLab.Script.Handlers
{
    //[HandlerName("Shares Initial")]
    [HandlerCategory(HandlerCategories.Position)]
    public class PositionShares : IPosition2Double
    {
        public double Execute(IPosition pos, int barNum)
        {
            if (pos == null || pos.EntryBarNum > barNum)
            {
                return 0;
            }
            return pos.SharesOrigin;
        }
    }

    //[HandlerName("Shares")]
    [HandlerCategory(HandlerCategories.Position)]
    public class PositionSharesByBar : IPosition2Double
    {
        public double Execute(IPosition pos, int barNum)
        {
            if (pos == null || pos.EntryBarNum > barNum)
            {
                return 0;
            }
            return pos.GetShares(barNum);
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    public class BalancedPrice : IPosition2Double
    {
        public double Execute(IPosition pos, int barNum)
        {
            if (pos == null || pos.EntryBarNum > barNum)
            {
                return 0;
            }
            return pos.GetBalancePrice(barNum);
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    public class EntryPrice : IPosition2Double
    {
        public double Execute(IPosition pos, int barNum)
        {
            if (pos == null || pos.EntryBarNum > barNum)
            {
                return 0;
            }
            return pos.EntryPrice;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    public class CalcEntryPrice : IPosition2Double
    {
        public double Execute(IPosition pos, int barNum)
        {
            return EntryPrice(pos, barNum);
        }

        public static double EntryPrice(IPosition pos, int barNum)
        {
            if (pos == null || pos.EntryBarNum > barNum)
            {
                return 0;
            }
            var posRt = pos as IPositionRt;
            if (posRt != null && posRt.EntryOrders.Any())
            {
                var order = posRt.EntryOrders.First();
                var orderPrice = order.OrderPrice;
                // ReSharper disable CompareOfFloatsByEqualityOperator
                var useOpen = order.OrderType == OrderType.Market || orderPrice == 0;
                // ReSharper restore CompareOfFloatsByEqualityOperator
                var entryPrice = pos.EntryBar.Open;
                useOpen |= order.OrderType == OrderType.Limit &&
                           (order.IsBuy && orderPrice > entryPrice || !order.IsBuy && orderPrice < entryPrice);
                return useOpen ? entryPrice : orderPrice;
            }
            return pos.EntryPrice;
        }
    }

    //[HandlerName("Entry Date")]
    [HandlerCategory(HandlerCategories.Position)]
    public class EntryDate : IPosition2Double
    {
        public double Execute(IPosition pos, int barNum)
        {
            if (pos == null || pos.EntryBarNum > barNum)
            {
                return 0;
            }
            var dt = pos.EntryBar.Date;
            return (dt.Year % 100) * 10000.0 + dt.Month * 100.0 + dt.Day;
        }
    }

    //[HandlerName("Entry Time")]
    [HandlerCategory(HandlerCategories.Position)]
    public class EntryTime : IPosition2Double
    {
        public double Execute(IPosition pos, int barNum)
        {
            if (pos == null || pos.EntryBarNum > barNum)
            {
                return 0;
            }
            var dt = pos.EntryBar.Date;
            return dt.Hour * 10000.0 + dt.Minute * 100.0 + dt.Second;
        }
    }

    //[HandlerName("Bars Held")]
    [HandlerCategory(HandlerCategories.Position)]
    public class BarsHeld : IPosition2Double
    {
        public double Execute(IPosition pos, int barNum)
        {
            if (pos == null || pos.EntryBarNum >= barNum)
            {
                return 0;
            }
            if (pos.IsActive || pos.ExitBarNum > barNum)
            {
                return barNum - pos.EntryBarNum;
            }
            return pos.ExitBarNum - pos.EntryBarNum;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    public class Profit : IPosition2Double
    {
        public double Execute(IPosition pos, int barNum)
        {
            if (pos == null)
            {
                return 0;
            }
            return pos.CurrentProfit(barNum);
        }
    }

    //[HandlerName("Profit %")]
    [HandlerCategory(HandlerCategories.Position)]
    public class ProfitPct : IPosition2Double
    {
        public double Execute(IPosition pos, int barNum)
        {
            if (pos == null)
            {
                return 0;
            }
            // ReSharper disable CompareOfFloatsByEqualityOperator
            return pos.Shares == 0 ? 0 : pos.CurrentProfitPct(barNum);
            // ReSharper restore CompareOfFloatsByEqualityOperator
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    public class MAE : IPosition2Double
    {
        public double Execute(IPosition pos, int barNum)
        {
            if (pos == null)
            {
                return 0;
            }
            return pos.OpenMAE(barNum);
        }
    }

    //[HandlerName("MAE %")]
    [HandlerCategory(HandlerCategories.Position)]
    public class MAEPct : IPosition2Double
    {
        public double Execute(IPosition pos, int barNum)
        {
            if (pos == null)
            {
                return 0;
            }
            return pos.OpenMAEPct(barNum);
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    public class MFE : IPosition2Double
    {
        public double Execute(IPosition pos, int barNum)
        {
            if (pos == null)
            {
                return 0;
            }
            return pos.OpenMFE(barNum);
        }
    }

    //[HandlerName("MFE %")]
    [HandlerCategory(HandlerCategories.Position)]
    public class MFEPct : IPosition2Double
    {
        public double Execute(IPosition pos, int barNum)
        {
            if (pos == null)
            {
                return 0;
            }
            return pos.OpenMFEPct(barNum);
        }
    }

    //[HandlerName("Trail Stop")]
    [HandlerCategory(HandlerCategories.Position)]
    public class TrailStop : IPosition2Double
    {
        [HandlerParameter(true, "1.5", Min = "0.1", Max = "5", Step = "0.1", Name = "Stop Loss")]
        public double StopLoss { get; set; }

        [HandlerParameter(true, "0.5", Min = "0.1", Max = "3", Step = "0.1", Name = "Trail Enable")]
        public double TrailEnable { get; set; }

        [HandlerParameter(true, "0.5", Min = "0.1", Max = "3", Step = "0.1", Name = "Trail Loss")]
        public double TrailLoss { get; set; }

        [HandlerParameter(true, "false", NotOptimized = true, Name = "Use Calc Price")]
        public bool UseCalcPrice { get; set; }

        public double Execute(IPosition pos, int barNum)
        {
            if (pos == null)
            {
                return 0;
            }
            double stop;
            var curProfit = pos.OpenMFE(barNum);
            var entryPrice = pos.EntryPrice;
            if (UseCalcPrice)
            {
                var ep2 = CalcEntryPrice.EntryPrice(pos, barNum);
                var diff = (entryPrice - ep2) * (pos.IsLong ? 1 : -1);
                curProfit += diff;
                entryPrice = ep2;
            }
            curProfit *= 100 / entryPrice * pos.Security.Margin;

            if (curProfit > TrailEnable)
            {
                double shift = (curProfit - TrailLoss) / 100;
                stop = entryPrice * (1 + (pos.IsLong ? shift : -shift));
            }
            else
            {
                double shift = (0 - StopLoss) / 100;
                stop = entryPrice * (1 + (pos.IsLong ? shift : -shift));
            }
            var lastStop = pos.GetStop(barNum);
            // ReSharper disable CompareOfFloatsByEqualityOperator
            if (lastStop == 0)
            // ReSharper restore CompareOfFloatsByEqualityOperator
            {
                return stop;
            }
            return pos.IsLong ? Math.Max(stop, lastStop) : Math.Min(stop, lastStop);
        }
    }

    //[HandlerName("Trail Stop Abs")]
    [HandlerCategory(HandlerCategories.Position)]
    public class TrailStopAbs : IPosition2Double
    {
        [HandlerParameter(true, "150", Min = "10", Max = "500", Step = "5", Name = "Stop Loss")]
        public double StopLoss { get; set; }

        [HandlerParameter(true, "50", Min = "10", Max = "500", Step = "5", Name = "Trail Enable")]
        public double TrailEnable { get; set; }

        [HandlerParameter(true, "50", Min = "10", Max = "500", Step = "5", Name = "Trail Loss")]
        public double TrailLoss { get; set; }

        [HandlerParameter(true, "false", NotOptimized = true, Name = "Use Calc Price")]
        public bool UseCalcPrice { get; set; }

        public double Execute(IPosition pos, int barNum)
        {
            if (pos == null)
            {
                return 0;
            }
            double stop;
            var curProfit = pos.OpenMFE(barNum);
            var entryPrice = pos.EntryPrice;
            if (UseCalcPrice)
            {
                var ep1 = entryPrice;
                var ep2 = CalcEntryPrice.EntryPrice(pos, barNum);
                var diff = (ep1 - ep2) * (pos.IsLong ? 1 : -1);
                curProfit += diff;
                entryPrice = ep2;
            }
            if (curProfit > TrailEnable)
            {
                double shift = curProfit - TrailLoss;
                stop = entryPrice + (pos.IsLong ? shift : -shift);
            }
            else
            {
                double shift = -StopLoss;
                stop = entryPrice + (pos.IsLong ? shift : -shift);
            }
            var lastStop = pos.GetStop(barNum);
            // ReSharper disable CompareOfFloatsByEqualityOperator
            if (lastStop == 0)
            // ReSharper restore CompareOfFloatsByEqualityOperator
            {
                return stop;
            }
            return pos.IsLong ? Math.Max(stop, lastStop) : Math.Min(stop, lastStop);
        }
    }

    //[HandlerName("Has Position Active")]
    [HandlerCategory(HandlerCategories.Position)]
    public class HasPositionActive : IBar2BoolHandler
    {
        public bool Execute(ISecurity source, int barNum)
        {
            var res = source.Positions.GetLastPositionActive(barNum) != null;
            if (res || !source.IsRealtime || source.SimulatePositionOrdering)
                return res;
            var lp = source.Positions.GetLastPosition(source.Positions.BarsCount);
            return lp != null && (lp.IsActive || lp.ExitBarNum > barNum);
        }
    }

    //[HandlerName("Has Long Position Active")]
    [HandlerCategory(HandlerCategories.Position)]
    public class HasLongPositionActive : IBar2BoolHandler
    {
        public bool Execute(ISecurity source, int barNum)
        {
            var res = source.Positions.GetLastLongPositionActive(barNum) != null;
            if (res || !source.IsRealtime || source.SimulatePositionOrdering)
                return res;
            var lp = source.Positions.GetLastLongPositionActive(source.Positions.BarsCount);
            if (lp != null)
                return true;
            lp = source.Positions.GetLastLongPositionClosed(source.Positions.BarsCount);
            return lp != null && lp.ExitBarNum > barNum;
        }
    }

    //[HandlerName("Has Short Position Active")]
    [HandlerCategory(HandlerCategories.Position)]
    public class HasShortPositionActive : IBar2BoolHandler
    {
        public bool Execute(ISecurity source, int barNum)
        {
            var res = source.Positions.GetLastShortPositionActive(barNum) != null;
            if (res || !source.IsRealtime || source.SimulatePositionOrdering)
                return res;
            var lp = source.Positions.GetLastShortPositionActive(source.Positions.BarsCount);
            if (lp != null)
                return true;
            lp = source.Positions.GetLastShortPositionClosed(source.Positions.BarsCount);
            return lp != null && lp.ExitBarNum > barNum;
        }
    }

    //[HandlerName("Last is Closed and was Long")]
    [HandlerCategory(HandlerCategories.Position)]
    public class LastClosedIsLong : IBar2BoolHandler
    {
        public bool Execute(ISecurity source, int barNum)
        {
            var closed = source.Positions.GetLastPosition(barNum);
            return closed != null && closed.IsLong && !closed.IsActiveForBar(barNum);
        }
    }

    //[HandlerName("Last is Closed and was Short")]
    [HandlerCategory(HandlerCategories.Position)]
    public class LastClosedIsShort : IBar2BoolHandler
    {
        public bool Execute(ISecurity source, int barNum)
        {
            var closed = source.Positions.GetLastPosition(barNum);
            return closed != null && closed.IsShort && !closed.IsActiveForBar(barNum);
        }
    }

    //[HandlerName("Last Closed Is Long")]
    [HandlerCategory(HandlerCategories.Position)]
    public class LastClosedIsLong2 : IBar2BoolHandler
    {
        public bool Execute(ISecurity source, int barNum)
        {
            var closed = source.Positions.GetLastPositionClosed(barNum);
            return closed != null && closed.IsLong;
        }
    }

    //[HandlerName("Last Closed Is Short")]
    [HandlerCategory(HandlerCategories.Position)]
    public class LastClosedIsShort2 : IBar2BoolHandler
    {
        public bool Execute(ISecurity source, int barNum)
        {
            var closed = source.Positions.GetLastPositionClosed(barNum);
            return closed != null && closed.IsShort;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    public sealed class PositionEntryBarNumber : IPosition2Double
    {
        public double Execute(IPosition position, int barNum)
        {
            return position?.EntryBarNum ?? -1;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    public sealed class PositionExitBarNumber : IPosition2Double
    {
        public double Execute(IPosition position, int barNum)
        {
            return position?.ExitBarNum ?? -1;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    public sealed class PositionChangeEntryBarNumber : IPosition2Double
    {
        public double Execute(IPosition position, int barNum)
        {
            if (position == null)
                return -1;

            var changeInfos = position.ChangeInfos;
            for (var i = changeInfos.Count - 1; i >= 0; i--)
            {
                var changeInfo = changeInfos[i];
                if (changeInfo.EntryBarNum >= 0)
                    return changeInfo.EntryBarNum;
            }
            return -1;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    public sealed class PositionChangeExitBarNumber : IPosition2Double
    {
        public double Execute(IPosition position, int barNum)
        {
            if (position == null)
                return -1;

            var changeInfos = position.ChangeInfos;
            for (var i = changeInfos.Count - 1; i >= 0; i--)
            {
                var changeInfo = changeInfos[i];
                if (changeInfo.ExitBarNum >= 0)
                    return changeInfo.ExitBarNum;
            }
            return -1;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    public sealed class PositionChangeEntryPrice : IPosition2Double
    {
        public double Execute(IPosition position, int barNum)
        {
            if (position == null)
                return double.NaN;

            var changeInfos = position.ChangeInfos;
            for (var i = changeInfos.Count - 1; i >= 0; i--)
            {
                var changeInfo = changeInfos[i];
                if (changeInfo.EntryBarNum >= 0)
                    return changeInfo.EntryPrice;
            }
            return double.NaN;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    public sealed class PositionChangeExitPrice : IPosition2Double
    {
        public double Execute(IPosition position, int barNum)
        {
            if (position == null)
                return double.NaN;

            var changeInfos = position.ChangeInfos;
            for (var i = changeInfos.Count - 1; i >= 0; i--)
            {
                var changeInfo = changeInfos[i];
                if (changeInfo.ExitBarNum >= 0)
                    return changeInfo.ExitPrice;
            }
            return double.NaN;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    public sealed class Median : IBar2ValueDoubleHandler
    {
        private int m_lastBarNum = -1;
        private double m_lastProfit;
        private double m_lastMaxProfit;

        public double Execute(ISecurity source, int barNum)
        {
            if (barNum - m_lastBarNum == 1)
            {
                m_lastProfit = source.GetProfit(barNum);
                if (m_lastMaxProfit < m_lastProfit)
                    m_lastMaxProfit = m_lastProfit;
            }
            else if (m_lastBarNum != barNum)
            {
                double profit = 0, maxProfit = double.NegativeInfinity;
                for (var i = 0; i <= barNum; i++)
                {
                    profit = source.GetProfit(i);
                    if (maxProfit < profit)
                        maxProfit = profit;
                }
                m_lastProfit = profit;
                m_lastMaxProfit = maxProfit;
            }
            m_lastBarNum = barNum;
            return m_lastMaxProfit - m_lastProfit;
        }
    }
}
// ReSharper restore MemberCanBePrivate.Global
// ReSharper restore UnusedAutoPropertyAccessor.Global
// ReSharper restore UnusedMember.Global
