﻿using System;
using System.Linq;

// ReSharper disable UnusedAutoPropertyAccessor.Global
// ReSharper disable MemberCanBePrivate.Global
// ReSharper disable UnusedMember.Global
namespace TSLab.Script.Handlers
{
    public abstract class ClosedPositionCache
    {
        private IPosition[] m_lastCache;
        private IPosition m_lastClosed;

        protected IPosition[] GetClosedPositions(ISecurity source, int barNum)
        {
            var last = source.Positions.GetLastPositionClosed(barNum);
            if (m_lastCache == null || last != m_lastClosed)
            {
                m_lastClosed = last;
                m_lastCache = source.Positions.GetClosedForBar(barNum).OrderByDescending(pos => pos.ExitBar.Date).ToArray();
            }
            return m_lastCache;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    public class DrowdownCount : ClosedPositionCache, IBar2ValueDoubleHandler
    {
        public double Execute(ISecurity source, int barNum)
        {
            return GetClosedPositions(source, barNum).TakeWhile(pos => pos.Profit() <= 0).Count();
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    public class HasTwoLoss : ClosedPositionCache, IBar2BoolHandler
    {
        public bool Execute(ISecurity source, int barNum)
        {
            var list = GetClosedPositions(source, barNum);
            if (list.Length < 2)
            {
                return false;
            }
            return list[0].Profit() < 0 && list[1].Profit() < 0;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    public sealed class IsItLossAtLastPosition : ClosedPositionCache, IBar2BoolHandler
    {
        public bool Execute(ISecurity source, int barNum)
        {
            var positions = GetClosedPositions(source, barNum);
            return positions.Length > 0 && positions[0].Profit() < 0;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    //[HandlerName("Last Exit has this name")]
    public class PosActiveNameExit : IBar2BoolHandler
    {
        [HandlerParameter(true, NotOptimized = true)]
        public string Name { get; set; }

        #region IBar2BoolHandler Members

        public bool Execute(ISecurity source, int barNum)
        {
            var pos = source.Positions.GetLastPositionClosed(barNum);
            string name = pos == null ? null : pos.ExitSignalName;
            return name == Name;
        }

        #endregion
    }

    [HandlerCategory(HandlerCategories.Position)]
    //[HandlerName("Last Closed Position Time")]
    public class LastClosedPositionTime : IBar2ValueDoubleHandler
    {
        public double Execute(ISecurity source, int barNum)
        {
            var pos = source.Positions.GetLastPositionClosed(barNum);
            var dt = pos == null ? DateTime.MinValue : pos.EntryBar.Date;
            return dt.Hour * 10000.0 + dt.Minute * 100.0 + dt.Second;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    //[HandlerName("Last Closed Position Date")]
    public class LastClosedPositionDate : IBar2ValueDoubleHandler
    {
        public double Execute(ISecurity source, int barNum)
        {
            var pos = source.Positions.GetLastPositionClosed(barNum);
            var dt = pos == null ? DateTime.MinValue : pos.EntryBar.Date;
            return (dt.Year % 100) * 10000.0 + dt.Month * 100.0 + dt.Day;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    //[HandlerName("Last Closed Position Exit Time")]
    public class LastClosedPositionExitTime : IBar2ValueDoubleHandler
    {
        public double Execute(ISecurity source, int barNum)
        {
            var pos = source.Positions.GetLastPositionClosed(barNum);
            var dt = pos == null ? DateTime.MinValue : pos.ExitBar.Date;
            return dt.Hour * 10000.0 + dt.Minute * 100.0 + dt.Second;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    //[HandlerName("Last Closed Position Exit Date")]
    public class LastClosedPositionExitDate : IBar2ValueDoubleHandler
    {
        public double Execute(ISecurity source, int barNum)
        {
            var pos = source.Positions.GetLastPositionClosed(barNum);
            var dt = pos == null ? DateTime.MinValue : pos.ExitBar.Date;
            return (dt.Year % 100) * 10000.0 + dt.Month * 100.0 + dt.Day;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    //[HandlerName("Last Closed Named Position Exit Time")]
    public class LastClosedNamePositionExitTime : IBar2ValueDoubleHandler
    {
        [HandlerParameter(true, NotOptimized = true)]
        public string Name { get; set; }

        public double Execute(ISecurity source, int barNum)
        {
            var pos = source.Positions.GetLastClosedForSignal(Name, barNum);
            if (pos != null && (pos.IsActive || pos.ExitBarNum > barNum))
            {
                pos = null;
            }
            var dt = pos == null ? DateTime.MinValue : pos.ExitBar.Date;
            return dt.Hour * 10000.0 + dt.Minute * 100.0 + dt.Second;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    //[HandlerName("Last Closed Named Position Exit Date")]
    public class LastClosedNamePositionExitDate : IBar2ValueDoubleHandler
    {
        [HandlerParameter(true, NotOptimized = true)]
        public string Name { get; set; }

        public double Execute(ISecurity source, int barNum)
        {
            var pos = source.Positions.GetLastClosedForSignal(Name, barNum);
            if (pos != null && (pos.IsActive || pos.ExitBarNum > barNum))
            {
                pos = null;
            }
            var dt = pos == null ? DateTime.MinValue : pos.ExitBar.Date;
            return (dt.Year % 100) * 10000.0 + dt.Month * 100.0 + dt.Day;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    //[HandlerName("Days In Position")]
    public class DaysInPosition : IPosition2Double
    {
        public double Execute(IPosition pos, int barNum)
        {
            if (pos == null)
            {
                return 0;
            }
            var dt = pos.EntryBar.Date;
            var curDt = pos.Security.Bars[barNum].Date;
            return (curDt - dt).TotalDays;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    //[HandlerName("Minutes In Position")]
    public class MinutesInPosition : IPosition2Double
    {
        public double Execute(IPosition pos, int barNum)
        {
            if (pos == null)
            {
                return 0;
            }
            var dt = pos.EntryBar.Date;
            var curDt = pos.Security.Bars[barNum].Date;
            return (curDt - dt).TotalMinutes;
        }
    }

    [HandlerCategory(HandlerCategories.Position)]
    //[HandlerName("Last Exit Price")]
    public class LastExitPrice : IBar2ValueDoubleHandler
    {
        public double Execute(ISecurity source, int barNum)
        {
            var pos = source.Positions.GetLastPositionClosed(barNum);
            return pos == null ? 0 : pos.ExitPrice;
        }
    }
}
