﻿using System;
using System.Collections.Generic;
using System.Linq;
using TSLab.DataSource;

// ReSharper disable UnusedAutoPropertyAccessor.Global
// ReSharper disable MemberCanBePrivate.Global
// ReSharper disable UnusedMember.Global
namespace TSLab.Script.Handlers
{
    public abstract class SecurityBase : IBar2DoubleHandler, IContextUses, IDoubleHandlerWithUpdate
    {
        protected abstract double GetData(IDataBar bar);

        protected virtual IEnumerable<double> GetData(IEnumerable<IDataBar> source)
        {
            return source.Select(GetData);
        }

        public IList<double> Execute(ISecurity source)
        {
            var data = GetData(source.Bars);
            if (Context != null && Context.IsOptimization)
            {
                return data.ToArray();
            }
            var size = (source.Bars.Count / BarUtils.ListSizeStep + 1) * BarUtils.ListSizeStep;
            var list = new List<double>(size);
            list.AddRange(data);
            return list;
        }

        public void Update(IList<double> data, IDataBar bar, bool isNewBar)
        {
            var ndt = GetData(bar);
            if (isNewBar)
            {
                data.Add(ndt);
            }
            else if (data.Count > 0)
            {
                data[data.Count - 1] = ndt;
            }
        }

        public IContext Context { get; set; }
    }

    [HandlerDecimals(0)]
    [HandlerCategory(HandlerCategories.TradeMath)]
    public class Volume : SecurityBase
    {
        protected override double GetData(IDataBar bar)
        {
            return bar.Volume;
        }

        protected override IEnumerable<double> GetData(IEnumerable<IDataBar> source)
        {
            return source.Select(b => b.Volume);
        }
    }

    [HandlerDecimals(0)]
    [HandlerCategory(HandlerCategories.TradeMath)]
    //[HandlerName("Open Interest")]
    public class OpenInterest : SecurityBase
    {
        protected override double GetData(IDataBar bar)
        {
            return bar.Interest;
        }

        protected override IEnumerable<double> GetData(IEnumerable<IDataBar> source)
        {
            return source.Select(b => b.Interest);
        }
    }

    [HandlerCategory(HandlerCategories.TradeMath)]
    public sealed class Time : IBar2DoubleHandler
    {
        public IList<double> Execute(ISecurity source)
        {
            return source.Bars.Select(b => b.Date.Hour * 10000.0 + b.Date.Minute * 100.0 + b.Date.Second).ToArray();
        }
    }

    //[HandlerName("Time In Mins")]
    [HandlerCategory(HandlerCategories.TradeMath)]
    public sealed class TimeInMins : IBar2DoubleHandler
    {
        public IList<double> Execute(ISecurity source)
        {
            return source.Bars.Select(b => b.Date.Hour * 60.0 + b.Date.Minute).ToArray();
        }
    }

    [HandlerCategory(HandlerCategories.TradeMath)]
    public sealed class Date : IBar2DoubleHandler
    {
        public IList<double> Execute(ISecurity source)
        {
            return source.Bars.Select(b => (b.Date.Year % 100) * 10000.0 + b.Date.Month * 100.0 + b.Date.Day).ToArray();
        }
    }

    //[HandlerName("Day Of Week")]
    [HandlerCategory(HandlerCategories.TradeMath)]
    public sealed class DayOfWeek : IBar2DoubleHandler
    {
        public IList<double> Execute(ISecurity source)
        {
            return source.Bars.Select(b => (b.Date.DayOfWeek == System.DayOfWeek.Sunday
                                                  ? 7.0
                                                  : (double)b.Date.DayOfWeek)).ToArray();
        }
    }

    //[HandlerName("Day Of Month")]
    [HandlerCategory(HandlerCategories.TradeMath)]
    public sealed class DayOfMonth : IBar2DoubleHandler
    {
        public IList<double> Execute(ISecurity source)
        {
            return source.Bars.Select(b => (double)b.Date.Day).ToArray();
        }
    }

    [HandlerCategory(HandlerCategories.TradeMath)]
    public class Open : IBar2DoubleHandler
    {
        public IList<double> Execute(ISecurity source)
        {
            return source.OpenPrices;
        }
    }

    [HandlerCategory(HandlerCategories.TradeMath)]
    public class Close : SecurityBase
    {
        protected override double GetData(IDataBar bar)
        {
            return bar.Close;
        }

        protected override IEnumerable<double> GetData(IEnumerable<IDataBar> source)
        {
            return source.Select(b => b.Close);
        }
    }

    [HandlerCategory(HandlerCategories.TradeMath)]
    public class Low : SecurityBase
    {
        protected override double GetData(IDataBar bar)
        {
            return bar.Low;
        }

        protected override IEnumerable<double> GetData(IEnumerable<IDataBar> source)
        {
            return source.Select(b => b.Low);
        }
    }

    [HandlerCategory(HandlerCategories.TradeMath)]
    public class High : SecurityBase
    {
        protected override double GetData(IDataBar bar)
        {
            return bar.High;
        }

        protected override IEnumerable<double> GetData(IEnumerable<IDataBar> source)
        {
            return source.Select(b => b.High);
        }
    }

    [HandlerCategory(HandlerCategories.TradeMath)]
    public class Bid : SecurityBase
    {
        protected override double GetData(IDataBar bar)
        {
            return bar is IBar ? ((IBar)bar).Bid : 0;
        }
    }

    [HandlerCategory(HandlerCategories.TradeMath)]
    public class Ask : SecurityBase
    {
        protected override double GetData(IDataBar bar)
        {
            return bar is IBar ? ((IBar)bar).Ask : 0;
        }
    }

    [HandlerCategory(HandlerCategories.TradeMath)]
    //[HandlerName("Step Price")]
    public class StepPrice : SecurityBase
    {
        protected override double GetData(IDataBar bar)
        {
            return bar is IBar ? ((IBar)bar).StepPrice : 0;
        }
    }

    [HandlerCategory(HandlerCategories.TradeMath)]
    //[HandlerName("Theoretical Price")]
    public class TheoreticalPrice : SecurityBase
    {
        protected override double GetData(IDataBar bar)
        {
            return bar is IBar ? ((IBar)bar).TheoreticalPrice : 0;
        }
    }

    [HandlerCategory(HandlerCategories.TradeMath)]
    //[HandlerName("Option Volatility")]
    public class OptVolatility : SecurityBase
    {
        protected override double GetData(IDataBar bar)
        {
            return bar is IBar ? ((IBar)bar).Volatility : 1;
        }
    }

    [HandlerCategory(HandlerCategories.TradeMath)]
    public class BidQty : SecurityBase
    {
        protected override double GetData(IDataBar bar)
        {
            return bar is IBar ? ((IBar)bar).BidQty : 0;
        }
    }

    [HandlerCategory(HandlerCategories.TradeMath)]
    public class AskQty : SecurityBase
    {
        protected override double GetData(IDataBar bar)
        {
            return bar is IBar ? ((IBar)bar).AskQty : 0;
        }
    }

    [HandlerCategory(HandlerCategories.TradeMath)]
    public class Cutter : IBar2BarHandler
    {
        [HandlerParameter(true, "2", NotOptimized = true)]
        public int Decimals
        {
            get;
            set;
        }

        public ISecurity Execute(ISecurity source)
        {
            var coef = Math.Pow(10, Decimals);
            var bars = source.Bars.Select(
                delegate(IDataBar bar)
                {
                    var babar = bar as IBar;
                    var nbar = babar == null
                        ? new DataBar()
                        : new BidAskBar
                        {
                            Ask = Math.Ceiling(babar.Ask * coef) / coef,
                            Bid = Math.Ceiling(babar.Bid * coef) / coef,
                            AskQty = babar.AskQty,
                            BidQty = babar.BidQty,
                            StepPrice = babar.StepPrice,
                            Volatility = babar.Volatility,
                            TheoreticalPrice = babar.TheoreticalPrice,
                        };
                    nbar.Date = bar.Date;
                    nbar.Open = Math.Ceiling(bar.Open * coef) / coef;
                    nbar.High = Math.Ceiling(bar.High * coef) / coef;
                    nbar.Low = Math.Ceiling(bar.Low * coef) / coef;
                    nbar.Close = Math.Ceiling(bar.Close * coef) / coef;
                    nbar.Volume = bar.Volume;
                    nbar.Interest = bar.Interest;
                    return nbar;
                });
            return source.CloneAndReplaceBars(bars);
        }
    }

    //[HandlerName("Multiply (Security)")]
    [HandlerCategory(HandlerCategories.TradeMath)]
    public class SecMultiply : IBar2BarHandler
    {
        [HandlerParameter(true, "10")]
        public double Coef
        {
            get;
            set;
        }

        public ISecurity Execute(ISecurity source)
        {
            var bars = source.Bars.Select(
                bar =>
                {
                    var babar = bar as IBar;
                    var nbar = babar == null
                        ? new DataBar()
                        : new BidAskBar
                        {
                            Ask = Math.Ceiling(babar.Ask * Coef),
                            Bid = Math.Ceiling(babar.Bid * Coef),
                            AskQty = babar.AskQty,
                            BidQty = babar.BidQty,
                            StepPrice = babar.StepPrice,
                            Volatility = babar.Volatility,
                            TheoreticalPrice = babar.TheoreticalPrice,
                        };
                    nbar.Date = bar.Date;
                    nbar.Open = Math.Ceiling(bar.Open * Coef);
                    nbar.High = Math.Ceiling(bar.High * Coef);
                    nbar.Low = Math.Ceiling(bar.Low * Coef);
                    nbar.Close = Math.Ceiling(bar.Close * Coef);
                    nbar.Volume = bar.Volume;
                    nbar.Interest = bar.Interest;
                    return nbar;
                });
            return source.CloneAndReplaceBars(bars);
        }
    }

    //[HandlerName("MultiplyWith")]
    [HandlerCategory(HandlerCategories.TradeMath)]
    public class SecMultiplyWith : IBarDouble2BarHandler
    {
        [HandlerParameter(true, "2", NotOptimized = true)]
        public int Decimals
        {
            get;
            set;
        }

        [HandlerParameter(true, "1")]
        public double Coef
        {
            get;
            set;
        }

        public ISecurity Execute(ISecurity source, IList<double> src2)
        {
            var dcf = Math.Pow(10, Decimals);
            var bars = new List<IDataBar>(source.Bars.Count);
            for (int i = 0; i < source.Bars.Count; i++)
            {
                var bar = source.Bars[i];
                var s2 = src2.Count == 0 ? 0 : src2[Math.Min(i, src2.Count - 1)];
                var coef = Coef * s2 * dcf;
                var babar = bar as IBar;
                var nbar = babar == null
                    ? new DataBar()
                    : new BidAskBar
                    {
                        Ask = Math.Ceiling(babar.Ask * coef) / dcf,
                        Bid = Math.Ceiling(babar.Bid * coef) / dcf,
                        AskQty = babar.AskQty,
                        BidQty = babar.BidQty,
                        StepPrice = babar.StepPrice,
                        Volatility = babar.Volatility,
                        TheoreticalPrice = babar.TheoreticalPrice,
                    };
                nbar.Date = bar.Date;
                nbar.Open = Math.Ceiling(bar.Open * coef) / dcf;
                nbar.High = Math.Ceiling(bar.High * coef) / dcf;
                nbar.Low = Math.Ceiling(bar.Low * coef) / dcf;
                nbar.Close = Math.Ceiling(bar.Close * coef) / dcf;
                nbar.Volume = bar.Volume;
                nbar.Interest = bar.Interest;
                bars.Add(nbar);
            }
            return source.CloneAndReplaceBars(bars);
        }
    }

    //[HandlerName("DivideWith")]
    [HandlerCategory(HandlerCategories.TradeMath)]
    public class SecDivideWith : IBarDouble2BarHandler
    {
        [HandlerParameter(true, "2", NotOptimized = true)]
        public int Decimals
        {
            get;
            set;
        }

        [HandlerParameter(true, "1")]
        public double Coef
        {
            get;
            set;
        }

        public ISecurity Execute(ISecurity source, IList<double> src2)
        {
            var dcf = Math.Pow(10, Decimals);
            var bars = new List<IDataBar>(source.Bars.Count);
            for (int i = 0; i < source.Bars.Count; i++)
            {
                var bar = source.Bars[i];
                var s2 = src2.Count == 0 ? 0 : src2[Math.Min(i, src2.Count - 1)];
                var coef = Coef * 1 / Math.Max(s2, 1e-10) * dcf;
                var babar = bar as IBar;
                var nbar = babar == null
                    ? new DataBar()
                    : new BidAskBar
                    {
                        Ask = Math.Ceiling(babar.Ask * coef) / dcf,
                        Bid = Math.Ceiling(babar.Bid * coef) / dcf,
                        AskQty = babar.AskQty,
                        BidQty = babar.BidQty,
                        StepPrice = babar.StepPrice,
                        Volatility = babar.Volatility,
                        TheoreticalPrice = babar.TheoreticalPrice,
                    };
                nbar.Date = bar.Date;
                nbar.Open = Math.Ceiling(bar.Open * coef) / dcf;
                nbar.High = Math.Ceiling(bar.High * coef) / dcf;
                nbar.Low = Math.Ceiling(bar.Low * coef) / dcf;
                nbar.Close = Math.Ceiling(bar.Close * coef) / dcf;
                nbar.Volume = bar.Volume;
                nbar.Interest = bar.Interest;
                bars.Add(nbar);
            }
            return source.CloneAndReplaceBars(bars);
        }
    }

    //[HandlerName("Absolute Commission")]
    [HandlerCategory(HandlerCategories.TradeMath)]
    [OutputsCount(0)]
    public class AbsolutCommission : ICommissionHandler
    {
        [HandlerParameter(true, "0.0002", NotOptimized = true, Editor = "CommissionTemplate")]
        public double Commission
        {
            get;
            set;
        }

        public void Execute(ISecurity source)
        {
            source.Commission = ComissionDelegate;
        }

        private double ComissionDelegate(IPosition pos, double price, double shares, bool isEntry, bool isPart)
        {
            return shares * Commission;
        }
    }

    //[HandlerName("Relative Commission")]
    [HandlerCategory(HandlerCategories.TradeMath)]
    [OutputsCount(0)]
    public class RelativeCommission : ICommissionHandler
    {
        [HandlerParameter(true, "0.05", NotOptimized = true, Editor = "CommissionPctTemplate")]
        public double CommissionPct
        {
            get;
            set;
        }

        [HandlerParameter(true, "10.0", NotOptimized = true)]
        public double MarginPct
        {
            get;
            set;
        }

        public void Execute(ISecurity source)
        {
            source.Commission = CommissionDelegate;
        }

        protected virtual double CommissionDelegate(IPosition pos, double price, double shares, bool isEntry, bool isPart)
        {
            var comm = price * CommissionPct / 100.0 * shares;
            if (!isEntry && !isPart)
            {
                var days = (pos.ExitBar.Date - pos.EntryBar.Date).TotalDays;
                var mshares = pos.MaxShares;
                if (pos.IsLong)
                {
                    var sh = pos.SharesOrigin;
                    if (sh <= 1)
                    {
                        mshares = 0;
                    }
                    else
                    {
                        mshares *= 1 - 1 / sh;
                    }
                }
                if (shares > 0)
                {
                    comm += pos.AverageEntryPrice * MarginPct / 100 * days / 365 * mshares;
                }
            }
            return comm;
        }
    }

    //[HandlerName("Relative Commission With Minimal")]
    [HandlerCategory(HandlerCategories.TradeMath)]
    [OutputsCount(0)]
    public class RelativeCommisionWithMinimal : RelativeCommission
    {
        [HandlerParameter(true, "30.0", NotOptimized = true)]
        public double MinimalCommission
        {
            get;
            set;
        }

        protected override double CommissionDelegate(IPosition pos, double price, double shares, bool isEntry, bool isPart)
        {
            var res = base.CommissionDelegate(pos, price, shares, isEntry, isPart);
            var minimalCommission = MinimalCommission / pos.Security.LotSize;
            if (res < minimalCommission)
            {
                res = minimalCommission;
            }
            return res;
        }
    }
}
// ReSharper restore MemberCanBePrivate.Global
// ReSharper restore UnusedAutoPropertyAccessor.Global
// ReSharper restore UnusedMember.Global
